/* Read & process config data */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>

#include "blocks.h"

#ifndef min
#define min(A,B) ((A)<(B)?(A):(B))
#endif

/* Read config data. Return OK on success */
int readcfg(FILE *fp)
{
    int i, tot;

    /* Read numeric data */
    if (readnums(fp) != OK) return !OK;

    /* Surround board with guard squares */
    bbwidth = (2*pmax - 2 + bwidth);
    bbheight = (2*pmax - 2 + bheight);

    /* Initialize the playing board */
    if (!(board = calloc(bbwidth * bbheight, 1))) return !OK;
    for(i=0; i<bheight; i++)
        memset(pbrd(0,i), EMPTY, bwidth);

    /* Read shape data */
    if (readshapes(fp) != OK) return !OK;

    /* Check square count */
    for(tot=i=0; i<smax; i++)
        tot += shapes[i].p;

    if ((tot -= bwidth * bheight) != 0)
    {
        if (tot<0)
            printe("Not enough pieces to cover area,"
                    " deficit=%d squares\n", -tot);
        else
            printe("Too many pieces to cover area,"
                    " excess=%d squares\n", tot);
        return !OK;
    }

    /* Show all shapes, in all rotations */
    if (verbose & 2)
        showshapes(shapes);

    return OK;
}

#define BUFLEN 80      /* (Should be enough :) */

/* Read next line from fp, skipping comments,
   which start with a ';' in column 1. Return NULL on error */
char *getline(FILE *fp)
{
    static char buf[BUFLEN];
    int len;

    do
    {
        if (!fgets(buf, BUFLEN, fp)) return NULL;
    } while (*buf == ';');

    /* Remove terminating \n, if present */
    len = strlen(buf);
    if (len>0 && buf[len-1]=='\n')
        buf[len-1] = '\0';

    return buf;
}

/* Read next line from fp, skipping blank lines & comments.
   Return NULL on error */
char *gettext(FILE *fp)
{
    char *buf;

    do
    {
        if (!(buf = getline(fp))) return NULL;
    } while (*buf == '\0');

    return buf;
}

typedef struct arg
{
    char *fmt;  /* Argument name & type, using scanf-style % format */
    void *ptr;  /* Pointer to data destination */
} ARG;

/* Read argument values from fp using args template */
int readargs(FILE *fp, ARG *args)
{
    int i, j, count, len;
    char *buf, *val;

    /* Count args */
    for (count=0; args[count].fmt; count++)
        ;

    /* Scan file for args */
    for(i=0; i < count; i++)
    {
        if (!(buf = gettext(fp))) return !OK;
        if (!(val = strchr(buf, '='))) return !OK;

        /* Get name length */
        len = val - buf;

        /* Search for this name in args */
        for(j=0; j < count; j++)
        {
            if (strncmp(args[j].fmt, buf, len) == 0)
                break;
        }
        if (j==count) return !OK;   /* Not found */

        /* Convert arg */
        if (sscanf(buf, args[j].fmt, args[j].ptr) != 1) return !OK;
    }

    return OK;
}

char *linetypes[] =
{
    "NONE", "ISO", "IBM1", "IBM2", "POV"
};

char *chartypes[] =
{
    "NONE", "ALPHA", "NUMERIC"
};

char *charsets[] =
{
    NULL,
    "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdef",
    "0123456789ABCDEFGHIJKLMNOPQRSTUV"
};

/* Input data buffers */
char linetype[BUFLEN],
    chartype[BUFLEN],
    colourtype[BUFLEN],
    dumptype[BUFLEN];

int readnums(FILE *fp)
{
    int i;
    ARG args[] =
    {
        {"SIZE=%d",     &pmax},
        {"NUMBER=%d",   &smax},
        {"WIDTH=%d",    &bwidth},
        {"HEIGHT=%d",   &bheight},
        {"COUNTER=%d",  &maxtest},
        {"SOLVE=%d",    &maxsolve},      
        {"LINES=%s",    linetype},
        {"CHARS=%s",    chartype},
        {"COLOUR=%s",   colourtype},
        {"DUMP=%s",     dumptype},

        {NULL, NULL}
    };

    if (readargs(fp, args) != OK) return !OK;

    if (smax > 32) return !OK;
    
    if (maxsolve < 1) maxsolve = 0;
    
    if (maxtest < 1) maxtest = 1;

    /* Convert linetype to numeric form */
    for (i=0; i<MAXL; i++)
        if (strcmp(linetype, linetypes[i]) == 0)
            break;
    if (i == MAXL) return !OK;
    lineflag = i;

    /* Convert chartype to numeric form */
    for (i=0; i<MAXC; i++)
        if (strcmp(chartype, chartypes[i]) == 0)
            break;

    if (i == MAXC)
        charset = chartype;     /* Use given string as charset */
    else
        charset = charsets[i];
    if (charset)                /* Stash length */
        charsetlen = strlen(charset);

    /* Convert colourtype to numeric form */
    colourflag = strcmp(colourtype, "TRUE")==0
               || strcmp(colourtype, "ON")==0
               || strcmp(colourtype, "YES")==0;

    /* Convert dumptype to numeric form */
    verbose = 0;
    if (strstr(dumptype, "TEXT")) verbose |= 1;
    if (strstr(dumptype, "GRAPHIC")) verbose |= 2;

    /* Choose display function */
    if (lineflag == POV)
        putboard = putboardPOV;
    else if (lineflag != NOLINE)
        putboard = putboardL;
    else if (charset==NULL && colourflag==0)
        putboard = putboardQ;
    else
        putboard = putboardNL;

#if 0
    printe("size=%d number=%d width=%d height=%d\n"
           "counter=%d solve=%d lines=<%s> chars=<%s>\n"
           "lineflag=%d charset=<%s>\n"
            "colourtype=<%s> colourflag=%d\n",
        pmax, smax, bwidth, bheight, maxtest, maxsolve, linetype, chartype,
        lineflag, charset, colourtype, colourflag);
#endif

    return OK;
}

int readshapes(FILE *fp)
{
    int i, k;
    char *buf;

    /* Allocate shapes structure array */
    if (!(shapes = calloc(smax, sizeof(SHAPE)))) return !OK;

    /* Allocate a shape char grid */
    if (!(grid = malloc(pmax * pmax))) return !OK;

    /* Read shapes */
    for (k=0; k<smax; k++)
    {
        /* Clear grid */
        memset(grid, EMPTY, pmax*pmax);

        /* Read shape lines */
        for(i=0; i<pmax; i++)
        {
            buf = i ? getline(fp) : gettext(fp);
            if (buf == NULL) return !OK;
            if (*buf == '\0') break;

            /* Copy line to grid */            
            memcpy(pgrid(grid, 0, i), buf, min(strlen(buf), pmax));
            //printe("%sBUF%d=[%s]\n", i?"":"\n", i, buf);
        }

        /* Build shape */
        if (grid2shape(grid, shapes+k) != OK)
        {
            printe("Bad shape %d\n", k);
                return !OK;
        }
        else if (verbose & 1)
            dumpshape(shapes+k);
    }

    return OK;
}

void freeshapes(SHAPE *shapes)
{
    int i;

    if (!shapes)
        return;

    for(i=0; i<smax; i++)
    {
        if (!shapes[i].cell)
            break;

        free(shapes[i].cell);
    }

    free(shapes);
}

/* Like printf, but prints to stderr */
int printe(char *fmt, ...)
{
  int rc;
  
  va_list args;
  
  va_start(args, fmt);
  rc = vfprintf(stderr, fmt, args);
  va_end(args);
  
  return rc;
}
